;;;
;;; Universal Monitor 6502
;;;   Copyright (C) 2019 Haruo Asano
;;;

	CPU	6502

TARGET:	EQU	"6502"

	INCLUDE	"config.inc"

	INCLUDE "../common.inc"

;;;
;;; ROM area
;;;

	ORG	ROM_B

CSTART:
	LDX	#low(STACK)
	TXS
	JSR	INIT

	LDA	#$00
	STA	DSADDR
	STA	DSADDR+1
	STA	SADDR
	STA	SADDR+1
	STA	GADDR
	STA	GADDR+1
	STA	PSPEC
	LDA	#'S'
	STA	HEXMOD

	;; Main loop
	LDA	#low(OPNMSG)
	STA	PT0
	LDA	#high(OPNMSG)
	STA	PT0+1
	JSR	STROUT

	IF USE_IDENT
	ENDIF

WSTART:
	LDA	#low(PROMPT)
	STA	PT0
	LDA	#high(PROMPT)
	STA	PT0+1
	JSR	STROUT
	JSR	GETLIN
	LDX	#0
	JSR	SKIPSP
	JSR	UPPER
	CMP	#0
	BEQ	WSTART

	CMP	#'D'
	BNE	M00
	JMP	DUMP
M00:
	CMP	#'G'
	BNE	M01
	JMP	GO
M01:
	CMP	#'S'
	BNE	M02
	JMP	SETM
M02:

ERR:
	LDA	#low(ERRMSG)
	STA	PT0
	LDA	#high(ERRMSG)
	STA	PT0+1
	JSR	STROUT
	JMP	WSTART

;;;
;;; Dump memory
;;;
DUMP:
	INX
	JSR	SKIPSP
	JSR	RDHEX
	LDA	CNT
	BNE	DP0

	;; No arg.
	JSR	SKIPSP
	LDA	INBUF,X
	BNE	ERR
DP00:	
	LDA	DSADDR
	CLC
	ADC	#128
	STA	DEADDR
	LDA	DSADDR+1
	ADC	#0
	STA	DEADDR+1
	JMP	DPM

	;; 1st arg. found
DP0:
	LDA	PT1
	STA	DSADDR
	LDA	PT1+1
	STA	DSADDR+1
	JSR	SKIPSP
	LDA	INBUF,X
	CMP	#','
	BEQ	DP1
	CMP	#0
	BNE	ERR
	;; No 2nd arg.
	JMP	DP00
DP1:
	INX
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LDA	CNT
	BEQ	ERR
	LDA	INBUF,X
	BNE	ERR
	LDA	PT1
	SEC
	ADC	#0
	STA	DEADDR
	LDA	PT1+1
	ADC	#0
	STA	DEADDR+1

	;; DUMP main
DPM:	
	LDA	DSADDR
	AND	#$F0
	STA	PT1
	LDA	DSADDR+1
	STA	PT1+1
	LDA	#0
	STA	DSTATE
DPM0:
	JSR	DPL
	LDA	PT1
	CLC
	ADC	#16
	STA	PT1
	LDA	PT1+1
	ADC	#0
	STA	PT1+1
	JSR	CONST
	BNE	DPM1
	LDA	DSTATE
	CMP	#2
	BCC	DPM0
	LDA	DEADDR
	STA	DSADDR
	LDA	DEADDR+1
	STA	DSADDR+1
	JMP	WSTART
DPM1:
	LDA	PT1
	STA	DSADDR
	LDA	PT1+1
	STA	DSADDR+1
	JMP	WSTART

	;; Dump line
DPL:
	LDA	PT1+1
	JSR	HEXOUT2
	LDA	PT1
	JSR	HEXOUT2
	LDA	#low(DSEP0)
	STA	PT0
	LDA	#high(DSEP0)
	STA	PT0+1
	JSR	STROUT
	LDX	#0
	LDY	#0
DPL0:
	JSR	DPB
	CPX	#16
	BNE	DPL0

	LDA	#low(DSEP1)
	STA	PT0
	LDA	#high(DSEP1)
	STA	PT0+1
	JSR	STROUT

	;; Print ASCII area
	LDX	#0
DPL1:
	LDA	INBUF,X
	CMP	#' '
	BCC	DPL2
	CMP	#$7F
	BCS	DPL2
	JSR	CONOUT
	JMP	DPL3
DPL2:
	LDA	#'.'
	JSR	CONOUT
DPL3:
	INX
	CPX	#16
	BNE	DPL1
	JMP	CRLF

	;; Dump byte
DPB:
	LDA	#' '
	JSR	CONOUT
	LDA	DSTATE
	BNE	DPB2
	;; Dump state 0
	TYA
	SEC
	SBC	DSADDR
	AND	#$0F
	BEQ	DPB1
	;; Still 0 or 2
DPB0:
	LDA	#' '
	STA	INBUF,X
	JSR	CONOUT
	LDA	#' '
	JSR	CONOUT
	INX
	INY
	RTS
	;; Found start address
DPB1:
	LDA	#1
	STA	DSTATE
DPB2:
	LDA	DSTATE
	CMP	#1
	BNE	DPB0
	;; Dump state 1
	LDA	(PT1),Y
	STA	INBUF,X
	JSR	HEXOUT2
	INX
	INY
	TYA
	CLC
	ADC	PT1
	STA	PT0
	LDA	PT1+1
	ADC	#0
	STA	PT0+1
	LDA	PT0
	CMP	DEADDR
	BNE	DPBE
	LDA	PT0+1
	CMP	DEADDR+1
	BNE	DPBE
	;; Found end address
	LDA	#2
	STA	DSTATE
DPBE:
	RTS

;;;
;;;  Go address
;;;
GO:
	INX
	JSR	SKIPSP
	JSR	RDHEX
	LDA	INBUF,X
	BEQ	G00
	JMP	ERR
G00:
	LDA	CNT
	BEQ	G0
	LDA	PT1
	STA	GADDR
	LDA	PT1+1
	STA	GADDR+1
G0:
	JMP	(GADDR)

;;;
;;; Set memory
;;;
SETM:
	INX
	JSR	SKIPSP
	JSR	RDHEX
	JSR	SKIPSP
	LDA	INBUF,X
	BEQ	SM0
	JMP	ERR
SM0:
	LDA	CNT
	BEQ	SM1
	LDA	PT1
	STA	SADDR
	LDA	PT1+1
	STA	SADDR+1
SM1:
	LDA	SADDR+1
	JSR	HEXOUT2
	LDA	SADDR
	JSR	HEXOUT2
	LDA	#low(DSEP1)
	STA	PT0
	LDA	#high(DSEP1)
	STA	PT0+1
	JSR	STROUT
	LDY	#0
	LDA	(SADDR),Y
	JSR	HEXOUT2
	LDA	#' '
	JSR	CONOUT
	JSR	GETLIN
	LDX	#0
	JSR	SKIPSP
	LDA	INBUF,X
	BNE	SM2
SM10:	
	;; Empty (Increment address)
	LDA	SADDR
	CLC
	ADC	#1
	STA	SADDR
	LDA	SADDR+1
	ADC	#0
	STA	SADDR+1
	JMP	SM1
SM2:
	CMP	#'-'
	BNE	SM3
	;; '-' (Decrement address)
	LDA	SADDR
	SEC
	SBC	#1
	STA	SADDR
	LDA	SADDR+1
	SBC	#0
	STA	SADDR+1
	JMP	SM1
SM3:
	CMP	#'.'
	BNE	SM4
	;; '.' (Quit)
	JMP	WSTART
SM4:
	JSR	RDHEX
	LDA	CNT
	BNE	SM40
	JMP	ERR
SM40:
	LDA	PT1
	LDY	#0
	STA	(SADDR),Y
	JMP	SM10

	;;
	;; Other support routines
	;;

STROUT:
	LDY	#0
STRO0:
	LDA	(PT0),Y
	BEQ	STROE
	JSR	CONOUT
	INY
	JMP	STRO0
STROE:
	RTS

HEXOUT2:
	PHA
	LSR	A
	LSR	A
	LSR	A
	LSR	A
	JSR	HEXOUT1
	PLA
HEXOUT1:
	AND	#$0F
	CLC
	ADC	#'0'
	CMP	#'9'+1
	BCC	HEXOUTE
	CLC
	ADC	#'A'-'9'-1
HEXOUTE:
	JMP	CONOUT

CRLF:
	LDA	#CR
	JSR	CONOUT
	LDA	#LF
	JMP	CONOUT

GETLIN:
	LDX	#0
GL0:
	JSR	CONIN
	CMP	#CR
	BEQ	GLE
	CMP	#LF
	BEQ	GLE
	CMP	#BS
	BEQ	GLB
	CMP	#DEL
	BEQ	GLB
	CMP	#' '
	BCC	GL0
	CMP	#$80
	BCS	GL0
	CPX	#BUFLEN-1
	BCS	GL0		; Too long
	STA	INBUF,X
	INX
	JSR	CONOUT
	JMP	GL0
GLB:
	CPX	#0
	BEQ	GL0
	DEX
	LDA	#BS
	JSR	CONOUT
	LDA	#' '
	JSR	CONOUT
	LDA	#BS
	JSR	CONOUT
	JMP	GL0
GLE:
	JSR	CRLF
	LDA	#0
	STA	INBUF,X
	RTS

SKIPSP:
	LDA	INBUF,X
	CMP	#' '
	BNE	SSE
	INX
	JMP	SKIPSP
SSE:
	RTS

UPPER:
	CMP	#'a'
	BCC	UPE
	CMP	#'z'+1
	BCS	UPE
	ADC	#'A'-'a'
UPE:
	RTS

RDHEX:
	LDA	#0
	STA	PT1
	STA	PT1+1
	STA	CNT
RH0:
	LDA	INBUF,X
	JSR	UPPER
	CMP	#'0'
	BCC	RHE
	CMP	#'9'+1
	BCC	RH1
	CMP	#'A'
	BCC	RHE
	CMP	#'F'+1
	BCS	RHE
	SEC
	SBC	#'A'-'9'-1
RH1:
	SEC
	SBC	#'0'
	ASL	PT1
	ROL	PT1+1
	ASL	PT1
	ROL	PT1+1
	ASL	PT1
	ROL	PT1+1
	ASL	PT1
	ROL	PT1+1
	CLC
	ADC	PT1
	STA	PT1
	INC	CNT
	INX
	JMP	RH0
RHE:
	RTS

OPNMSG:
	FCB	CR,LF,"Universal Monitor 6502",CR,LF,$00
PROMPT:
	FCB	"] ",$00
IHEMSG:
	FCB	"Error ihex",CR,LF,$00

SHEMSG:
	FCB	"Error srec",CR,LF,$00

ERRMSG:
	FCB	"Error",CR,LF,$00

DSEP0:
	FCB	" :",$00
DSEP1:
	FCB	" : ",$00
IHEXER:
        FCB	":00000001FF",CR,LF,$00
SRECER:
        FCB	"S9030000FC",CR,LF,$00

	IF	USE_DEV_6551
	INCLUDE	"dev/dev_6551.asm"
	ENDIF

	IF	USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF

	;;
	;; Entry point
	;;

	ORG	ENTRY+0		; Cold start
E_CSTART:
	JMP	CSTART

	ORG	ENTRY+8		; Warm start
E_WSTART:
	JMP	WSTART

	ORG	ENTRY+16	; Console output
E_CONOUT:
	JMP	CONOUT

	ORG	ENTRY+24	; (Console) String output
E_STROUT:
	JMP	STROUT

	ORG	ENTRY+32	; Console input
E_CONIN:
	JMP	CONIN

	ORG	ENTRY+40	; Console status
E_CONST:
	JMP	CONST

	;;
	;; Vector area
	;; 

	ORG	$FFFA

	FDB	$0000		; IRQ/BRK

	FDB	CSTART		; RESET

	FDB	$0000		; NMI

;;;
;;; RAM	area
;;;

	;;
	;; Work Area
	;;

	ORG	WORK_B

INBUF:	RMB	BUFLEN		; Line input buffer
DSADDR:	RMB	2		; Dump start address
DEADDR:	RMB	2		; Dump end address
DSTATE:	RMB	1		; Dump state
GADDR:	RMB	2		; Go address
SADDR:	RMB	2		; Set address
HEXMOD:	RMB	1		; HEX file mode
RECTYP:	RMB	1		; Record type
PSPEC:	RMB	1		; Processor spec.

DMPPT:	RMB	2
CKSUM:	RMB	1		; Checksum

PT0:	RMB	2		; Generic Pointer 0
PT1:	RMB	2		; Generic Pointer 1
CNT:	RMB	1

	END
	
